--[[-------------------------------------------------------------------
--  ProcRateInformer - Igor Szyporyn Jørgensen
-------------------------------------------------------------------]]--
local hooksecurefunc, select, UnitGUID, tonumber, strfind, math, tonumber
  = hooksecurefunc, select, UnitGUID, tonumber, strfind, math, tonumber

local addonName, addon = ...;
local L = addon.L;

local craftingCastIteration = 0;
local craftingCastsForCompletion = 0;
local craftingInProgress = false;
local craftingTableItem = nil;
local craftingYield = 0;
local itemsInBagsPost = 0;
local itemsInBagsPre = 0;

local cachedCrafted = nil;
local cachedExpected = nil;
local cachedHistoricalProcRate = nil;
local cachedHistoricallyExpected = nil;
local cachedItemLink = nil;
local cachedProcRate = nil;
local cachedProcs = nil;

--[[-------------------------------------------------------------------
--  Utility functions
-------------------------------------------------------------------]]--
local function countItems(itemID)
  local c = 0;
  for bag=0,NUM_BAG_SLOTS do
      for slot=1,GetContainerNumSlots(bag) do
          if itemID == GetContainerItemID(bag,slot) then
              c=c+(select(2,GetContainerItemInfo(bag,slot)))
          end
      end
  end
  return c
end

local function round(num, numDecimalPlaces)
  local mult = 10^(numDecimalPlaces or 0)
  return math.floor(num * mult + 0.5) / mult
end

local function getItemStringFromItemLink(itemLink)
  local itemString = string.match(itemLink, "item[%-?%d:]+");

  return itemString;
end

local function getItemIdFromLink(itemLink)
  local returnValue = nil;
  if itemLink then
    local _, itemID = strsplit(":", string.match(itemLink, "item[%-?%d:]+"));

    if itemID then
      returnValue = tonumber(itemID);
    end
  end

  return returnValue;
end

local function getTableItemFromLink(itemLink)
  local tableItem = nil;

  if not (itemLink == nil or itemLink == "") then
    local itemID = getItemIdFromLink(itemLink);

    if itemID then
      for key, value in pairs(ProcRateInformerTable) do
        if value.itemID == itemID then
          tableItem = value;
        end 
      end
    end
  end

  return tableItem;
end

local function getTableItemFromId(id)

end

--[[-------------------------------------------------------------------
--  Print Report functions
-------------------------------------------------------------------]]--
local function printSessionReport(expectedCount, craftedCount, procs, craftedProcRate, historicallyExpected, historicalProcRate, itemLink)
  addon:Print(L["Session report for %s"], itemLink);
  addon:Print(L["Proc Rate: %s (%s crafts + %s procs = %s total)"], round(craftedProcRate, 2), expectedCount, procs, craftedCount);
  addon:Print(L["Historical Proc Rate: %s (%s crafts)"], round(historicalProcRate, 2), historicallyExpected);
end

local function printLastSessionReport()
  if cachedExpected then
    printSessionReport(cachedExpected, cachedCrafted, cachedProcs, cachedProcRate, cachedHistoricallyExpected, cachedHistoricalProcRate, cachedItemLink);
  else
    addon:Print(L["No last crafting session recorded"]);
  end
end

local function printHistoricalReport(tableItem)
  addon:Print(L["%s Proc Rate: %s (%s crafts)"], tableItem.name, round((tableItem.crafted / tableItem.expected), 2), tableItem.expected);
end


--[[-------------------------------------------------------------------
--  Finalize the craft calculations, emit procs and set historical values
-------------------------------------------------------------------]]--
function addon:FinalizeCraft()

  if craftingInProgress then
    local recipeID = craftingTableItem.ID;
    local itemID = craftingTableItem.itemID;  

    itemsInBagsPost = countItems(itemID);

    local _, itemLink, _, _, _, _, _, _, _, _, _ = GetItemInfo(itemID);
    local craftedCount = itemsInBagsPost - itemsInBagsPre;
    local expectedCount = craftingCastIteration * craftingYield;
    local procs = craftedCount - expectedCount;
    local historicalProcRate = nil;

    if not ProcRateInformerTable[recipeID].crafted then
      ProcRateInformerTable[recipeID].crafted = craftedCount;
      ProcRateInformerTable[recipeID].expected = expectedCount;
      ProcRateInformerTable[recipeID].procs = procs;
    else
      local newHistoricalCrafted = ProcRateInformerTable[recipeID].crafted + craftedCount;
      local newHistoricalExpected = ProcRateInformerTable[recipeID].expected + expectedCount;
      local newHistoricalProcs = ProcRateInformerTable[recipeID].procs + procs;

      ProcRateInformerTable[recipeID].crafted = newHistoricalCrafted;
      ProcRateInformerTable[recipeID].expected = newHistoricalExpected;
      ProcRateInformerTable[recipeID].procs = newHistoricalProcs;

      historicalProcRate = newHistoricalCrafted / newHistoricalExpected;
    end

    local historicallyCrafted = ProcRateInformerTable[recipeID].crafted;
    local historicallyExpected = ProcRateInformerTable[recipeID].expected;
    local historicalProcs = ProcRateInformerTable[recipeID].procs;

    if procs > 0 then
      local craftedProcRate = craftedCount / expectedCount;

      if historicalProcRate == nil then
        historicalProcRate = craftedProcRate;
      end
  
      cachedCrafted = craftedCount;
      cachedExpected = expectedCount;
      cachedHistoricalProcRate = historicalProcRate;
      cachedHistoricallyExpected = historicallyExpected;
      cachedItemLink = itemLink;
      cachedProcRate = craftedProcRate;
      cachedProcs = procs;

      printSessionReport(expectedCount, craftedCount, procs, craftedProcRate, historicallyExpected, historicalProcRate, itemLink);
    else
      self:Print(L["No procs recorded for your craft of %s"], itemLink);
    end

    self:InitProperties();
  end
end


--[[-------------------------------------------------------------------
--  Cast event handlers (craft start, each successfull cast, cast aborts)
-------------------------------------------------------------------]]--

local function onCraftStart(recipeID, amount)
  craftingTableItem = ProcRateInformerTable[recipeID];

  if craftingTableItem then
    craftingYield = C_TradeSkillUI.GetRecipeNumItemsProduced(recipeID);
    craftingInProgress = true;
    craftingCastIteration = 0;
    craftingCastsForCompletion = amount / craftingYield;
    itemsInBagsPre = countItems(craftingTableItem.itemID);
  end
end

function addon:OnSpellCastSuccess()
  if craftingInProgress then
    craftingCastIteration = craftingCastIteration + 1;
    if craftingCastIteration == craftingCastsForCompletion then
      self:Wait(1.5, function()
        addon:FinalizeCraft();
      end)
    end
  end
end

function addon:OnSpellCastFailed()
  if craftingInProgress then
    self:Wait(1.5, function()
      addon:FinalizeCraft();
    end)
  end
end

function addon:InitProperties()
  craftingCastIteration = 0;
  craftingCastsForCompletion = 0;
  craftingInProgress = false;
  craftingTableItem = nil;
  craftingYield = 0;
  itemsInBagsPost = 0;
  itemsInBagsPre = 0;
end


--[[-------------------------------------------------------------------
--  Initializing methods
-------------------------------------------------------------------]]--
function addon:initGlobals()
  if ProcRateInformerConfig == nil then
    ProcRateInformerConfig = {
      verbose = false
    };
  end

  if ProcRateInformerTable == nil then
    ProcRateInformerTable = {};
  end
end

function addon:InitVersion()
  ProcRateInformerConfig.version = 082;
  ProcRateInformerConfig.versionString = "0.8.2";
end

function addon:LegacyCleanUp()
  ProcRateInformerHistory = nil;
end

function addon:Enable()
  addon:Print(ProcRateInformerConfig.versionString);
end


--[[-------------------------------------------------------------------
--  Command handler functions.
-------------------------------------------------------------------]]--

--[[-------------------------------------------------------------------
--  Shows special requested report for either an item or last session
--  /pri show -> show all table data
--  /pri show last -> shows last session data
--  /pri show [Ancient Mana Potion] -> shows data for an item (requires item link!)
-------------------------------------------------------------------]]--
local function showReportCmd(options)
  if not (options == nil or options == "") then
    if options == "last" then printLastSessionReport();
    elseif options == "all" then
      local hadValueToShow = false;
      for key, value in pairs(ProcRateInformerTable) do
        if value.crafted then
          hadValueToShow = true;
          printHistoricalReport(value);
        end
      end

      if not hadValueToShow then
        addon:Print(L["No historical data"]);
      end
    else
      local item = getTableItemFromLink(options);

      if item then
        local itemLink = options;

        if item and item.crafted then
          printHistoricalReport(item);
        else
          addon:Print(L["No historical data for %s"], itemLink);
        end
      end
    end
  else
    addon:Print(L["/pri show all - Show all recorded data"]);
    addon:Print(L["/pri show last - Show last recorded session data"]);
    addon:Print(L["/pri show [link] - Show recorded data for item link"]);
  end
end


--[[-------------------------------------------------------------------
--  Resets data for all or requested item
--  /pri reset all -> will reset all historical data
--  /pri reset [Ancient Mana Potion] - will reset historical data for item (requires item link!)
--  /pri reset hard -> Will reset all table definitions
-------------------------------------------------------------------]]--
local function resetProcRateInformerTable()
  ProcRateInformerTable = nil;
  addon:Print(L["ProcRateInformerTable is rest - reloading UI"]);
  ReloadUI();
end

local function resetCmd(option)
  if not (option == nil or option == "") then
    if option == "all" then
      for key, value in pairs(ProcRateInformerTable) do
        if value.crafted then
          ProcRateInformerTable[value.ID].crafted = nil;
          ProcRateInformerTable[value.ID].expected = nil;
          ProcRateInformerTable[value.ID].procs = nil;
          addon:Print(L["All recorded data has been reset"], value.name);
        end 
      end
    elseif option == "hard" then
      StaticPopup_Show("DIALOG_PROCRATEINFORMER");
    else
      local item = getTableItemFromLink(option);

      if item then
        local tableItem = ProcRateInformerTable[item.ID];
        if tableItem then
          ProcRateInformerTable[item.ID].crafted = nil;
          ProcRateInformerTable[item.ID].expected = nil;
          ProcRateInformerTable[item.ID].procs = nil;
          addon:Print(L["Recorded data for %s has been reset"], item.name);
        else
          addon:Print(L["No table entry found to reset %s"], item.name);
        end
      end
    end
  else
    addon:Print(L["/pri reset all - Reset all recorded data"]);
    addon:Print(L["/pri reset [link] - Reset recorded data for item link"]);
    addon:Print(L["/pri reset hard - Reset all table definitions and data"]);
  end
end


--[[-------------------------------------------------------------------
--  Shows version
-------------------------------------------------------------------]]--
local function showVersion()
  addon:Print(ProcRateInformerConfig.versionString);
end



--[[-------------------------------------------------------------------
--  Called to handle commands.
-------------------------------------------------------------------]]--
local function showHelp()
  addon:Print(L["/pri show - Show help for listing recorded data"]);
  addon:Print(L["/pri reset - Show help for resetting data and recipe table"]);
  addon:Print(L["/pri version - Show current version"]);
end

local function slashCommandHandler(params, editBox)

  local _, _, command, options = string.find(params,"([%w%p]+)%s*(.*)$");

  if (command) then
    command = string.lower(command);
  end

  if (command == nil or command == "") then
    showHelp();
  elseif (command == "reset") then resetCmd(options);
  elseif (command == "version") then showVersion();
  elseif (command == "show") then showReportCmd(options);
  elseif (command == "help") then showHelp();
  else
    showHelp();
  end
end


StaticPopupDialogs["DIALOG_PROCRATEINFORMER"] = {
  text = L["Do you wish to clear the recipe table and reload?"],
  button1 = L["Yes"],
  button2 = L["No"],
  OnAccept = function()
    resetProcRateInformerTable();
  end,
  timeout = 0,
  whileDead = true,
  hideOnEscape = true,
  preferredIndex = 3
};

SlashCmdList["PROCRATEINFORMER"] = slashCommandHandler;
SLASH_PROCRATEINFORMER1 = "/pri";


--[[-------------------------------------------------------------------
--  Addon Initialization
-------------------------------------------------------------------]]--
function addon:Initialize()
  self:initGlobals();
  self:InitVersion();
  self:InitProperties();
  self:LegacyCleanUp();

  hooksecurefunc(C_TradeSkillUI, "CraftRecipe", onCraftStart);
  self:RegisterEvent("UNIT_SPELLCAST_SUCCEEDED", "OnSpellCastSuccess");
  self:RegisterEvent("UNIT_SPELLCAST_INTERRUPTED", "OnSpellCastFailed");
  self:RegisterEvent("UNIT_SPELLCAST_FAILED", "OnSpellCastFailed");
end